/*

Copyright (c) 2009-2018, Thanos Theo. All rights reserved.
Released Under a Simplified BSD (FreeBSD) License
for academic, personal & non-commercial use.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

The views and conclusions contained in the software and documentation are those
of the author and should not be interpreted as representing official policies,
either expressed or implied, of the FreeBSD Project.

A Commercial License is also available for commercial use with
special restrictions and obligations at a one-off fee. See links at:
1. http://www.dotredconsultancy.com/openglrenderingenginetoolrelease.php
2. http://www.dotredconsultancy.com/openglrenderingenginetoolsourcecodelicence.php
Please contact Thanos Theo (thanos.theo@dotredconsultancy.com) for more information.

*/

#pragma once

#ifndef CUDA_DEVICE_SCOPE_GUARD_H
#define CUDA_DEVICE_SCOPE_GUARD_H

#include "CUDAUtilityFunctions.h"

/** @brief namespace UtilsCUDA for encapsulating all the CUDA related code compiled by the NVCC compiler.
*
* @author Thanos Theo, 2018
* @version 14.0.0.0
*/
namespace UtilsCUDA
{
  /** @brief A lightweight and efficient class that sets the device for the current scope and restores the device that was originally set once scope ends in RAII style.
  *
  * @author Elias Stehle, 2020
  * @version 14.0.0.0
  */
  class CUDADeviceScopeGuard final
  {
  public:

    explicit CUDADeviceScopeGuard(int device) noexcept
    {
      restoreDevice_ = false;
      // for invalid device ids do not apply changes (e.g., device < 0 means 'leave device unchanged')
      if (device >= 0)
      {
        // get the device currently set
        CUDAError_checkCUDAError(cudaGetDevice(&originalDevice_));
        // override the currently set device to the one requested by the user
        if (originalDevice_ != device)
        {
          restoreDevice_ = true;
          CUDAError_checkCUDAError(cudaSetDevice(device));
        }
      }
    }

    ~CUDADeviceScopeGuard() noexcept // no virtual destructor for data-oriented design (no up-casting should ever be used)
    {
      if (restoreDevice_)
      {
        CUDAError_checkCUDAError(cudaSetDevice(originalDevice_));
      }
    }

    CUDADeviceScopeGuard(const CUDADeviceScopeGuard&) = delete; // copy-constructor deleted
    CUDADeviceScopeGuard(CUDADeviceScopeGuard&&)      = delete; // move-constructor deleted
    CUDADeviceScopeGuard& operator=(const CUDADeviceScopeGuard&) = delete; //      assignment operator deleted
    CUDADeviceScopeGuard& operator=(CUDADeviceScopeGuard&&)      = delete; // move-assignment operator deleted

  private:

    bool restoreDevice_  = false;
    int  originalDevice_ = 0;
  };
} // namespace UtilsCUDA

#endif // CUDA_DEVICE_SCOPE_GUARD_H